part of masamune_location;

final locationProvider = ChangeNotifierProvider((_) => LocationModel());

/// This class manages and monitors location information.
///
/// Execute [listen()] to request location permission and monitor for changes.
class LocationModel extends ValueModel<LocationCompassData> {
  LocationModel()
      : super(
          LocationCompassData._(
            location: LocationData.fromMap({
              'latitude': 0.0,
              'longitude': 0.0,
              'accuracy': 0.0,
              'altitude': 0.0,
              'speed': 0.0,
              'speed_accuracy': 0.0,
              'heading': 0.0,
              'time': 0.0,
            }),
            compass: CompassData.fromList([]),
          ),
        );

  // ignore: cancel_subscriptions
  StreamSubscription<CompassEvent>? _compassEventSubscription;
  // ignore: cancel_subscriptions
  StreamSubscription<LocationData>? _locationEventSubscription;

  /// Returns itself after the load/save finishes.
  Future<void>? get future => _completer?.future;
  Completer<void>? _completer;

  @protected
  final Location _location = Location();

  /// Check whether the location is currently being acquired.
  bool get isListened {
    if (!isPermitted) {
      return false;
    }
    if (_compassEventSubscription == null ||
        _locationEventSubscription == null) {
      return false;
    }
    return true;
  }

  /// Check the permission status of the current location.
  bool get isPermitted => _permissionStatus == PermissionStatus.granted;
  PermissionStatus _permissionStatus = PermissionStatus.denied;

  /// This class manages and monitors location information.
  ///
  /// Execute [listen()] to request location permission and monitor for changes.
  Future<void> listen({
    Duration timeout = const Duration(seconds: 60),
  }) async {
    await _listenLocation(timeout);
  }

  /// Abandon location information acquisition.
  void unlisten() {
    dispose();
  }

  Future<void> _listenLocation(Duration timeout) async {
    if (isListened || _completer != null) {
      return future;
    }
    _completer = Completer<void>();
    try {
      if (!await _location.serviceEnabled().timeout(timeout)) {
        if (!await _location.requestService().timeout(timeout)) {
          throw Exception(
            "Location service not available. The platform may not be supported or it may be disabled in the settings. please confirm."
                .localize(),
          );
        }
      }
      _permissionStatus = await _location.hasPermission().timeout(timeout);
      if (_permissionStatus == PermissionStatus.denied) {
        _permissionStatus =
            await _location.requestPermission().timeout(timeout);
      }
      if (_permissionStatus != PermissionStatus.granted) {
        throw Exception(
          "You are not authorized to use the location information service. Check the permission settings."
              .localize(),
        );
      }
      value = value.copyWith(
          location: await _location.getLocation().timeout(timeout));
      _compassEventSubscription = FlutterCompass.events?.listen((compass) {
        value = value.copyWith(compass: compass);
      });
      _locationEventSubscription =
          _location.onLocationChanged.listen((location) {
        value = value.copyWith(location: location);
      });
      notifyListeners();
    } catch (e) {
      _completer?.completeError(e);
      _completer = null;
      rethrow;
    } finally {
      _completer?.complete();
      _completer = null;
    }
  }

  /// Destroys the object.
  ///
  /// Destroyed objects are not allowed.
  @override
  void dispose() {
    _completer?.complete();
    _completer = null;
    _compassEventSubscription?.cancel();
    _locationEventSubscription?.cancel();
    _compassEventSubscription = null;
    _locationEventSubscription = null;
    super.dispose();
  }
}

/// Class for saving location information and compass information.
@immutable
class LocationCompassData {
  /// Class for saving location information and compass information.
  const LocationCompassData._({required this.location, required this.compass});

  /// Location data.
  final LocationData location;

  /// Compass data.
  final CompassEvent compass;

  @override
  bool operator ==(Object other) => other.hashCode == hashCode;

  @override
  int get hashCode => location.hashCode ^ compass.hashCode;

  LocationCompassData copyWith(
      {LocationData? location, CompassEvent? compass}) {
    return LocationCompassData._(
      location: location ?? this.location,
      compass: compass ?? this.compass,
    );
  }
}

/// Class that stores compass data.
typedef CompassData = CompassEvent;
